import gzip
import json
import logging
import re

try:
    from importlib import resources
except ImportError:
    import importlib_resources as resources  # type: ignore

from typing import Dict, List, Optional, Union
from wsgiref import simple_server

from webob import exc
from webob import Request
from webob import Response
from webob import static

from pyloot.backends.base import BaseBackend
from pyloot.backends.memory import InMemoryBackend
from pyloot.types import ObjectDescriptor

logger = logging.getLogger(__name__)

ResponseData = Union[List, Dict]


def _get_param(req: Request, name: str, default=None, typ=None):
    val = req.GET.get(name)
    if val:
        if typ is not None:
            val = typ(val)
    else:
        val = default
    return val


class PyLootServer:
    def __init__(
        self, backend: Optional[BaseBackend] = None, disable_response_gzip: bool = False
    ):
        self._handlers = {
            ("GET", "/static/[a-z0-9._-]+"): self._static,
            ("GET", "/api/history"): self._get_history,
            ("GET", "/api/objects"): self._get_objects,
            ("POST", "/api/objects"): self._post_objects,
            ("GET", "/api/objects/([0-9]+)"): self._get_object_by_id,
            ("GET", "/api/objects/([0-9]+)/children"): self._get_object_children,
            ("GET", "/api/objects/([0-9]+)/parents"): self._get_object_parents,
        }
        if backend:
            self._storage = backend
        else:
            self._storage = InMemoryBackend()

        self._disable_response_gzip: bool = disable_response_gzip

    def get_backend(self) -> BaseBackend:
        return self._storage

    def _make_response(
        self, data: ResponseData, status: Optional[int] = None
    ) -> Response:
        response = Response(json=data, status=status)
        if not self._disable_response_gzip:
            response.encode_content()
        return response

    def __call__(self, environ, start_response):
        if isinstance(self._storage, InMemoryBackend):
            assert not environ["wsgi.multiprocess"], (
                "pyloot middleware is not usable in a " "multi-process environment"
            )
        return self._dispatch(environ, start_response)

    def _dispatch(self, environ, start_response):
        req = Request(environ)
        logger.info(
            "[_dispatch] path_info=%s path=%s",
            req.path_info,
            req.path,
        )

        if req.path_info == "/" and req.method == "GET":
            logger.info(
                "[_dispatch] %s handler=%s", req.path_info, self._static.__name__
            )
            return self._static(req)(environ, start_response)

        for handler_spec, handler in self._handlers.items():
            method, handler_pattern = handler_spec
            match = re.fullmatch(handler_pattern, req.path_info.rstrip("/"))
            if method == req.method and match:
                logger.info(
                    "[_dispatch] %s handler=%s", req.path_info, handler.__name__
                )
                return handler(req, *match.groups())(environ, start_response)
        return exc.HTTPNotFound("No endpoint found")(environ, start_response)

    def _get_history(self, req: Request) -> Response:
        top = _get_param(req, "top", typ=int)
        items = self._storage.fetch_history(top=top)
        data = [item._asdict() for item in items]
        logger.info("[history] returning %s items [top=%s]", len(data), top)
        return self._make_response(data)

    def _get_objects(self, req: Request) -> Response:
        limit = _get_param(req, "limit", typ=int)
        group = _get_param(req, "group")
        if group:
            items = self._storage.fetch_by_group(group, limit=limit)
        else:
            items = self._storage.fetch(limit=limit)
        data = [item._asdict() for item in items]
        logger.info("[objects] returning %s items [group=%s]", len(data), group)
        return self._make_response(data)

    def _post_objects(self, req: Request) -> Response:
        data = gzip.GzipFile(fileobj=req.body_file).read().decode(req.charset)
        items = json.loads(data)
        items = [ObjectDescriptor(**item) for item in items]
        self._storage.store(items)
        logger.info("[objects] stored %s items", len(items))
        return self._make_response({})

    def _get_object_by_id(self, _: Request, _id: str) -> Response:
        item = self._storage.fetch_by_id(int(_id))
        status = None
        if item:
            data = item._asdict()
            logger.info("[objects_by_id] found %s [%s]", _id, item.type_name)
        else:
            logger.error("[objects_by_id] not found %s", _id)
            status = 404
            data = dict(error="Item not found")

        return self._make_response(data, status=status)

    def _get_object_children(self, _: Request, _id: str) -> Response:
        items = self._storage.fetch_children_of(int(_id))
        data = [item._asdict() for item in items]
        logger.info("[object children] returning %s items [id=%s]", len(data), _id)

        return self._make_response(data)

    def _get_object_parents(self, _: Request, _id: str) -> Response:
        items = self._storage.fetch_parents_of(int(_id))
        data = [item._asdict() for item in items]
        logger.info("[object parents] returning %s items [id=%s]", len(data), _id)

        return self._make_response(data)

    def _static(self, req: Request):
        """Static path where images and other files live"""
        if req.path_info_peek() == "static":
            req.path_info_pop()
        logger.info("[static asset] %s", req.path_info)
        try:
            with resources.files("pyloot").joinpath("static") as path:
                return static.DirectoryApp(path)
        except AttributeError:
            with resources.path("pyloot", "static") as path:
                return static.DirectoryApp(path)

    def serve_forever(self, host: str = "0.0.0.0", port: int = 8000):
        httpd = simple_server.make_server(host, port, self)
        try:
            logger.info("Starting PyLoot Server at http://%s:%s", host, port)
            httpd.serve_forever()
        except KeyboardInterrupt:
            logger.info("Shutting down")
            httpd.server_close()



import logging
import time
from collections import defaultdict
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from pyloot.backends.base import BaseBackend
from pyloot.types import ObjectDescriptor
from pyloot.types import ObjectTypeHistory


logger = logging.getLogger(__name__)


class InMemoryBackend(BaseBackend):
    def __init__(self, max_history: int = 300):
        self._data: Dict[int, ObjectDescriptor] = {}
        self._history: Dict[Tuple[str, str], List[int]] = defaultdict(list)
        self._sample_size: int = 0
        self._max_history: int = max_history

    def store(self, object_data: List[ObjectDescriptor]):
        cur_time = time.time()
        type_counts: Dict[Tuple[str, str], int] = defaultdict(int)
        to_delete = set(self._data.keys())
        for i, descr in enumerate(object_data):
            if descr.id in self._data:
                object_data[i] = descr._replace(seen=self._data[descr.id].seen)
                to_delete.remove(descr.id)
            else:
                object_data[i] = descr._replace(seen=cur_time)

            type_counts[(descr.type_module, descr.type_name)] += 1

            del descr

        for descr_tuple, count in type_counts.items():
            if descr_tuple not in self._history:
                self._history[descr_tuple] = [0] * self._sample_size
            self._history[descr_tuple].append(count)

            if len(self._history[descr_tuple]) > self._max_history:
                self._history[descr_tuple].pop(0)

        del type_counts

        if self._sample_size < self._max_history:
            self._sample_size += 1

        for _id in to_delete:
            del self._data[_id]

        del to_delete

        for descr in object_data:
            self._data[descr.id] = descr

        del object_data

    def fetch(self, limit: Optional[int] = None) -> List[ObjectDescriptor]:
        results = list(self._data.values())
        if limit is None or limit <= 0:
            return results
        return results[:limit]

    def fetch_by_id(self, _id: int) -> Optional[ObjectDescriptor]:
        return self._data.get(_id)

    def fetch_children_of(self, _id: int) -> List[ObjectDescriptor]:
        item = self.fetch_by_id(_id)
        if not item:
            logger.warning("Unable to find item with id %s", _id)
            return []

        items = [self._data.get(idx) for idx in item.child_ids]
        return [item for item in items if item]

    def fetch_parents_of(self, _id: int) -> List[ObjectDescriptor]:
        item = self.fetch_by_id(_id)
        if not item:
            logger.warning("Unable to find item with id %s", _id)
            return []

        items = [self._data.get(idx) for idx in item.parent_ids]
        return [item for item in items if item]

    def fetch_by_group(
        self, group: str, limit: Optional[int] = None
    ) -> List[ObjectDescriptor]:
        items = [item for item in self._data.values() if item.group == group]
        sorted_items = sorted(items, key=lambda item: item.seen or item.id)
        if limit is None or limit <= 0:
            return sorted_items
        return sorted_items[:limit]

    def fetch_history(self, top: Optional[int] = None) -> List[ObjectTypeHistory]:
        history = [
            self._calculate_history(descr_tuple[0], descr_tuple[1], counts)
            for descr_tuple, counts in self._history.items()
        ]
        sorted_history = sorted(history, reverse=True, key=lambda h: h.counts[-1])
        if top is None or top <= 0:
            return sorted_history
        return sorted_history[:top]

    def get_ids(self) -> Iterable[int]:
        yield id(self)
        yield id(self._data)
        yield id(self._history)
        yield id(self.__dict__)

        for descr_tuple, counts in self._history.items():
            yield id(descr_tuple)
            yield id(counts)

        for descr in self._data.values():
            yield id(descr)
            yield id(descr.parent_ids)
            yield id(descr.child_ids)
            yield id(descr.attrs)

    @classmethod
    def _calculate_history(
        cls, type_module: str, type_name: str, counts: List[int]
    ) -> ObjectTypeHistory:
        _min = min(counts)
        _max = max(counts)
        return ObjectTypeHistory(
            type_name=type_name,
            type_module=type_module,
            counts=counts,
            min=_min,
            max=_max,
        )



import cgi
import gc
import logging
import sys
from collections import defaultdict
from types import FrameType
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set

from pyloot.types import ObjectDescriptor

logger = logging.getLogger(__name__)

METHOD_TYPES = [
    type(tuple.__le__),  # 'wrapper_descriptor'
    type([1].__le__),  # 'method-wrapper'
    type(sys.getswitchinterval),  # 'builtin_function_or_method'
    type(cgi.FieldStorage.getfirst),  # 'instancemethod'
    type(logger.addFilter),  # 'bound instancemethod'
]


def _type_name(obj: object) -> str:
    return type(obj).__name__


def _safe_repr(obj: object) -> str:
    try:
        return repr(obj)
    except Exception as e:
        return "(__repr_error__:{}:{})".format(_type_name(e), _type_name(obj))


def _safe_getattr(obj: object, k: str, default: Any = Exception):
    try:
        from zope.interface.ro import C3

        if isinstance(obj, C3) or issubclass(obj, C3):  # type: ignore
            if k.startswith("ORIG_"):
                return "__ignored_zope_interface_C3_{}__".format(k)
    except (ImportError, TypeError):
        pass

    try:
        return getattr(obj, k)
    except Exception as e:
        if default is Exception:
            return "__getattr_error__:{}".format(_type_name(e))
        else:
            return default


def _get_pretty_name(obj: object) -> str:
    obj_type = type(obj)
    type_name = "%s.%s" % (obj_type.__module__, obj_type.__name__)
    pretty_type = type_name.replace("__builtin__.", "")
    pretty_type = pretty_type.replace("builtins.", "")

    name = _safe_getattr(obj, "__name__", default=None)
    if name is not None:
        pretty_type = "%s %r" % (pretty_type, name)
    return pretty_type


def _safe_get_attrs(obj: object) -> Dict[str, str]:
    try:
        keys = dir(obj)
    except Exception as e:
        return {"__dir_error__": "{}:{}".format(_type_name(e), str(e))}
    kv_pairs = [(k, _safe_getattr(obj, k)) for k in keys]
    return {k: _safe_repr(v) for k, v in kv_pairs if type(v) not in METHOD_TYPES}


def _should_include_object(ref: object, ignore_set: Set[int]) -> bool:
    thisfile = sys._getframe().f_code.co_filename
    # Exclude all frames that are from this module.
    if isinstance(ref, FrameType) and ref.f_code.co_filename == thisfile:
        return False

    # Exclude all functions and classes from this module or reftree.
    mod = str(getattr(ref, "__module__", ""))
    if "pyloot" in mod or mod == "__main__":
        return False  # pragma: nocover -- avoid bug in coverage due to Python's peephole optimizer

    if id(ref) in ignore_set:
        return False

    return True


def get_child_ids(obj: object) -> List[int]:
    """
    Return children of the provided object using gc.get_referents

    :param obj: The object
    :return: List of object ids
    """
    return [id(child) for child in gc.get_referents(obj)]


def get_data(obj: object) -> ObjectDescriptor:
    """
    Return a Object descriptor for the given object

    :param obj: The object
    :return:
    """
    obj_type = type(obj)
    return ObjectDescriptor(
        type_name=str(obj_type.__name__),
        type_module=str(obj_type.__module__),
        obj_name=_get_pretty_name(obj),
        id=id(obj),
        attrs=_safe_get_attrs(obj),
        parent_ids=[],
        child_ids=get_child_ids(obj),
        repr=_safe_repr(obj),
    )


def get_object_descriptors(
    ignored: Optional[Iterable[int]] = None,
) -> List[ObjectDescriptor]:
    """
    Return list of ::class::`ObjectDescriptor` instances for all objects in memory
    after a call to gc.collect.

    WARNING: Logging in this thread can cause a race condition if gevent is enabled
    """

    if ignored:
        ignore_set = set(ignored)
    else:
        ignore_set = set()

    ignore_set.add(id(ignore_set))

    del ignored

    gc.collect()

    objs: List[object] = gc.get_objects()

    objs = [obj for obj in objs if _should_include_object(obj, ignore_set)]
    results = [get_data(obj) for obj in objs]
    del objs
    del ignore_set
    child_to_parent: Dict[int, Set[int]] = defaultdict(set)

    for descr in results:
        for child_id in descr.child_ids:
            child_to_parent[child_id].add(descr.id)

    for descr in results:
        if descr.id in child_to_parent:
            descr.parent_ids.extend(child_to_parent[descr.id])

    del child_to_parent

    return results



import gzip
import http.client
import io
import json
import logging
import urllib.error
import urllib.request
from typing import cast
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from pyloot.backends.base import BaseBackend
from pyloot.types import ObjectDescriptor
from pyloot.types import ObjectTypeHistory


logger = logging.getLogger(__name__)


class HTTPRemoteBackend(BaseBackend):
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        prefix: Optional[str] = None,
    ):
        self._host = host or "localhost"
        self._port = port or 8000
        self._prefix = prefix or "/"

    def _make_url(self, endpoint: Optional[str] = None) -> str:
        endpoint = endpoint or "/"
        return "http://{host}:{port}{prefix}/{endpoint}".format(
            host=self._host,
            port=self._port,
            prefix=self._prefix.rstrip("/"),
            endpoint=endpoint.lstrip("/"),
        )

    @staticmethod
    def compress(entity: str) -> bytes:
        out = io.BytesIO()
        with gzip.GzipFile(fileobj=out, mode="w") as f:
            f.write(entity.encode("utf-8"))
        return out.getvalue()

    @staticmethod
    def decompress(entity: bytes) -> str:
        buf = io.BytesIO(entity)
        f = gzip.GzipFile(fileobj=buf)
        return f.read().decode()

    @classmethod
    def _read_response(cls, response: http.client.HTTPResponse) -> str:
        if response.info().get("Content-Encoding") == "gzip":
            return cls.decompress(response.read())
        else:
            return response.read().decode()

    @classmethod
    def _read_error(cls, response: urllib.error.HTTPError) -> str:
        if response.headers.get("Content-Encoding", "identity") == "gzip":
            return cls.decompress(response.read())
        else:
            return response.read().decode()

    def _make_request(
        self,
        endpoint: Optional[str] = None,
        method: Optional[str] = None,
        json_data: Optional[Union[Dict, List]] = None,
    ) -> urllib.request.Request:
        headers = {
            "Accept-Encoding": "gzip",
        }

        url = self._make_url(endpoint)
        data = None
        if json_data:
            if not method:
                method = "POST"

            data = self.compress(json.dumps(json_data))

            headers["Content-Encoding"] = "gzip"
            headers["Content-Type"] = "application/json"

        method = method or "GET"
        return urllib.request.Request(url, data=data, method=method, headers=headers)

    def _request_json(self, request: urllib.request.Request) -> Union[Dict, List]:
        response: http.client.HTTPResponse = urllib.request.urlopen(request, timeout=5)
        data = self._read_response(response)
        if response.getheader("Content-Type") == "application/json":
            return json.loads(data)

        return dict(content=data)

    def store(self, object_data: List[ObjectDescriptor]):
        data = [dict(item._asdict()) for item in object_data]
        for item in data:
            try:
                json.dumps(item)
            except:
                logger.exception("Unable to dump %s", item)
        request = self._make_request("/api/objects", json_data=data)
        try:
            self._request_json(request)
        except:
            logger.exception("Error storing data")

    def fetch(self, limit: Optional[int] = None) -> List[ObjectDescriptor]:
        if limit is not None:
            request = self._make_request("/api/objects?limit={}".format(limit))
        else:
            request = self._make_request("/api/objects")

        try:
            data = cast(List, self._request_json(request))
        except:
            logger.exception("Error fetching data")
            raise

        return [ObjectDescriptor(**item) for item in data]

    def fetch_by_id(self, _id: int) -> Optional[ObjectDescriptor]:
        request = self._make_request("/api/objects/{}".format(_id))

        try:
            data = cast(Dict, self._request_json(request))
        except urllib.error.HTTPError as e:
            if e.code == 404:
                logger.warning("Unable to find object with id %s", _id)
                return None
            else:
                logger.exception("Error fetching data")
                raise
        except:
            logger.exception("Error fetching data")
            raise

        return ObjectDescriptor(**data)

    def fetch_children_of(self, _id: int) -> List[ObjectDescriptor]:
        request = self._make_request("/api/objects/{}/children".format(_id))
        try:
            data = cast(Dict, self._request_json(request))
        except urllib.error.HTTPError as e:
            if e.code == 404:
                logger.warning("Unable to find object with id %s", _id)
                return []
            else:
                logger.exception("Error fetching data")
                raise
        except:
            logger.exception("Error fetching data")
            raise

        return [ObjectDescriptor(**item) for item in data]

    def fetch_parents_of(self, _id: int) -> List[ObjectDescriptor]:
        request = self._make_request("/api/objects/{}/parents".format(_id))
        try:
            data = cast(Dict, self._request_json(request))
        except urllib.error.HTTPError as e:
            if e.code == 404:
                logger.warning("Unable to find object with id %s", _id)
                return []
            else:
                logger.exception("Error fetching data")
                raise
        except:
            logger.exception("Error fetching data")
            raise

        return [ObjectDescriptor(**item) for item in data]

    def fetch_by_group(
        self, group: str, limit: Optional[int] = None
    ) -> List[ObjectDescriptor]:
        if limit is not None:
            request = self._make_request(
                "/api/objects?group={}&limit={}".format(group, limit)
            )
        else:
            request = self._make_request("/api/objects?group={}".format(group))
        try:
            data = cast(List, self._request_json(request))
        except:
            logger.exception("Error fetching data")
            raise

        return [ObjectDescriptor(**item) for item in data]

    def fetch_history(self, top: Optional[int] = None) -> List[ObjectTypeHistory]:
        if top is not None:
            request = self._make_request("/api/history?top={}".format(top))
        else:
            request = self._make_request("/api/history")
        try:
            data = cast(List, self._request_json(request))
        except:
            logger.exception("Error fetching data")
            raise

        return [ObjectTypeHistory(**item) for item in data]

    def get_ids(self) -> Iterable[int]:
        yield id(self)



from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional


class ObjectDescriptor(NamedTuple):
    repr: str
    type_name: str
    type_module: str
    obj_name: str
    id: int
    attrs: Dict[str, str]
    parent_ids: List[int]
    child_ids: List[int]
    seen: Optional[float] = None

    @property
    def group(self):
        return "{}.{}".format(self.type_module, self.type_name)


class ObjectTypeHistory(NamedTuple):
    type_name: str
    type_module: str
    counts: List[int]
    min: int
    max: int
